import uuid
import json
import requests
from .MercariItemFull import Item as MercariItemFull, ItemAuction
from .DpopUtils import generate_DPOP

rootURL = "https://api.mercari.jp/"
rootProductURL = "https://jp.mercari.com/item/"
searchURL = "{}v2/entities:search".format(rootURL)
itemInfoURL = "{}items/get".format(rootURL) # idk why not v2


class MercariSearchStatus:
    DEFAULT = "STATUS_DEFAULT"
    ON_SALE = "STATUS_ON_SALE"
    SOLD_OUT = "STATUS_SOLD_OUT"

class MercariSort:
    SORT_DEFAULT = 'SORT_DEFAULT'
    SORT_CREATED_TIME = 'SORT_CREATED_TIME'
    SORT_NUM_LIKES = 'SORT_NUM_LIKES'
    SORT_SCORE = 'SORT_SCORE'
    SORT_PRICE = 'SORT_PRICE'

class MercariOrder:
    ORDER_DESC = 'ORDER_DESC'
    ORDER_ASC = 'ORDER_ASC'

class MercariItemStatus:
    ITEM_STATUS_UNSPECIFIED = 'ITEM_STATUS_UNSPECIFIED'
    ITEM_STATUS_ON_SALE = 'ITEM_STATUS_ON_SALE'
    ITEM_STATUS_TRADING = 'ITEM_STATUS_TRADING'
    ITEM_STATUS_SOLD_OUT = 'ITEM_STATUS_SOLD_OUT'
    ITEM_STATUS_STOP = 'ITEM_STATUS_STOP'
    ITEM_STATUS_CANCEL = 'ITEM_STATUS_CANCEL'
    ITEM_STATUS_ADMIN_CANCEL = 'ITEM_STATUS_ADMIN_CANCEL'

class Item:
    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.productURL = "{}{}".format(rootProductURL, kwargs['id'])
        self.imageURL = kwargs['thumbnails'][0]
        self.productName = kwargs['name']
        self.price = kwargs['price']
        self.status = kwargs['status']
        self.soldOut = kwargs['status'] != MercariItemStatus.ITEM_STATUS_SOLD_OUT
        self.created = kwargs['created']
        self.updated = kwargs['updated']
        # this is optional, only present if the item is an auction
        if "auction" in kwargs and kwargs["auction"] is not None:
            self.auction = ItemAuction(**kwargs['auction'])
        else:
            self.auction = None

    @staticmethod
    def fromApiResp(apiResp):
        return Item(
            **apiResp
        )

# because requests is doing some dumb bullshit and using capital booleans
# we'll force lowercase booleans to fix this dumb shit
def convert_booleans(obj):
    if isinstance(obj, bool):
        return str(obj).lower()
    elif isinstance(obj, dict):
        # Recursively process each key-value pair in the dictionary
        return {k: convert_booleans(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        # Recursively process each item in the list
        return [convert_booleans(item) for item in obj]
    else:
        return obj
    
# used for the search endpoint
# returns [] if resp has no items on it
# returns [Item's] otherwise
def parse(resp):
    if(len(resp["items"]) == 0):
        return [], False, None

    respItems = resp["items"]
    nextPageToken = resp["meta"]["nextPageToken"]
    return [Item.fromApiResp(item) for item in respItems], bool(nextPageToken), nextPageToken

# used for the itemInfo endpoint
def parseItemInfo(resp):
    return MercariItemFull(
        **resp['data']
    )

def fetch(url, data, parser, method="POST"):
    DPOP = generate_DPOP(
        # let's see if this gets blacklisted, but it also lets them track
        uuid="Mercari Python Bot",
        method=method,
        url=url,
    )

    headers = {
        'DPOP': DPOP,
        'X-Platform': 'web',  # mercari requires this header
        'Accept': '*/*',
        'Accept-Encoding': 'deflate, gzip',
        'Content-Type': 'application/json; charset=utf-8',
        # courtesy header since they're blocking python-requests (returns 0 results)
        'User-Agent': 'python-mercari',
    }
    
    serializedData = json.dumps(data, ensure_ascii=False).encode('utf-8')

    if method == "POST":
        r = requests.post(url, headers=headers, data=serializedData)
    else:
        r = requests.get(url, headers=headers, params=convert_booleans(data))    

    r.raise_for_status()

    return parser(r.json())

# not sure if the v1 prefix ever changes, but from quick testing, doesn't seem like it
def pageToPageToken(page):
    return "v1:{}".format(page)

# returns an generator for Item objects
# keeps searching until no results so may take a while to get results back

def search(keywords, sort=MercariSort.SORT_CREATED_TIME, order=MercariOrder.ORDER_DESC, status=MercariSearchStatus.ON_SALE, exclude_keywords=""):

    # This is per page and not for the final result
    limit = 120

    data = {
        # this seems to be random, but we'll add a prefix for mercari to track if they wanted to
        "userId": "MERCARI_BOT_{}".format(uuid.uuid4()), 
        "pageSize": limit,
        "pageToken": pageToPageToken(0),
        # same thing as userId, courtesy of a prefix for mercari
        "searchSessionId": "MERCARI_BOT_{}".format(uuid.uuid4()),
        # this is hardcoded in their frontend currently, so leaving it
        "indexRouting": "INDEX_ROUTING_UNSPECIFIED",
        "searchCondition": {
            "keyword": keywords,
            "sort": sort,
            "order": order,
            "status": [status],
            "excludeKeyword": exclude_keywords,
        },
        "withAuction": True,
        # I'm not certain what these are, but I believe it's what mercari queries against
        # this is the default in their site, so leaving it as these 2
        "defaultDatasets": [
            "DATASET_TYPE_MERCARI",
            "DATASET_TYPE_BEYOND"
        ]
    }

    has_next_page = True

    while has_next_page:
        items, has_next_page, next_page_token = fetch(searchURL, data, parse)
        yield from items
        data['pageToken'] = next_page_token


def getItemInfo(itemID, country_code="US"):
    data = {
        "id": itemID,
        "country_code": country_code,
        "include_item_attributes": True,
        "include_product_page_component": True,
        "include_non_ui_item_attributes": True,
        "include_donation": True,
        "include_offer_like_coupon_display": True,
        "include_offer_coupon_display": True,
        "include_item_attributes_sections": True,
        "include_auction": True
    }

    item = fetch(itemInfoURL, data, parseItemInfo, method="GET")
    return item


# some endpoints use similar keys but different case, bidDeadline vs bid_deadline
def get_from_kwargs(kwargs, *keys, default=""):
    for key in keys:
        if key in kwargs:
            return kwargs[key]
    return default

class Printable:
    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(f'{k}={v!r}' for k, v in self.__dict__.items())})"

    def __str__(self, level=0):
        indent = '\t' * level
        items = []
        for k, v in self.__dict__.items():
            if isinstance(v, Printable):
                items.append(f"{indent}\"{k}\": {v.__str__(level + 1)}")
            elif isinstance(v, str):
                items.append(f"{indent}\"{k}\": \"{v}\"")
            elif isinstance(v, list) and all(isinstance(i, Printable) for i in v):
                if len(v) == 0:
                    items.append(f"{indent}\"{k}\": []")
                else:
                    nested_items = f", ".join(i.__str__(level + 2) for i in v)
                    items.append(f"{indent}\"{k}\": [ {nested_items}\n{indent}\t]")
            else:
                items.append(f"{indent}\"{k}\": {v}")
        return '{\n\t' + ',\n\t'.join(items) + f'\n{indent}}}'

class Ratings(Printable):
    good: int
    normal: int
    bad: int

    def __init__(self, *args, **kwargs):
        self.good = kwargs['good']
        self.normal = kwargs['normal']
        self.bad = kwargs['bad']

class ConvertedPrice(Printable):
    price: int
    currency_code: str
    rate_updated: int

    def __init__(self, *args, **kwargs):
        self.price = kwargs['price']
        self.currency_code = kwargs['currency_code']
        self.rate_updated = kwargs['rate_updated']

class Seller(Printable):
    id: str
    name: str
    photo_url: str
    photo_thumbnail_url: str
    created: int
    num_sell_items: int
    ratings: Ratings
    num_ratings: int
    score: int
    is_official: bool
    quick_shipper: bool
    is_followable: bool
    is_blocked: bool
    star_rating_score: int

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']
        self.photo_url = kwargs['photo_url']
        self.photo_thumbnail_url = kwargs['photo_thumbnail_url']
        self.created = kwargs['created']
        self.num_sell_items = kwargs['num_sell_items']
        self.ratings = Ratings(**kwargs['ratings'])
        self.num_ratings = kwargs['num_ratings']
        self.score = kwargs['score']
        self.is_official = kwargs['is_official']
        self.quick_shipper = kwargs['quick_shipper']
        self.is_followable = kwargs['is_followable']
        self.is_blocked = kwargs['is_blocked']
        self.star_rating_score = kwargs['star_rating_score']

class ItemCategory(Printable):
    id: str
    name: str
    display_order: int
    parent_category_id: int
    parent_category_name: str
    root_category_id: int
    root_category_name: str
    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']
        self.display_order = kwargs['display_order']
        self.parent_category_id = kwargs['parent_category_id']
        self.parent_category_name = kwargs['parent_category_name']
        self.root_category_id = kwargs['root_category_id']
        self.root_category_name = kwargs['root_category_name']

class ItemCategoryNTiers(ItemCategory):
    brand_group_id: int = 0

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        if 'brand_group_id' in kwargs:
          self.brand_group_id = kwargs['brand_group_id']

class ParentCategoryNTier(Printable):
    id: str
    name: str
    display_order: int

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']
        self.display_order = kwargs['display_order']

class ItemCondition(Printable):
    id: str
    name: str

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']

class ShippingPayer(Printable):
    
    id: str
    name: str
    code: str

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']
        self.code = kwargs['code']

class ShippingMethod(Printable):
    id: str
    name: str
    is_deprecated: bool

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']
        self.is_deprecated = kwargs['is_deprecated']

class ShippingFromArea(Printable):
    id: str
    name: str

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']

class ShippingDuration(Printable):
    id: str
    name: str
    min_days: int
    max_days: int

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']
        self.min_days = kwargs['min_days']
        self.max_days = kwargs['max_days']

class ShippingClass(Printable):
    id: str
    fee: int
    icon_id: int
    pickup_fee: int
    shipping_fee: int
    total_fee: int
    is_pickup: bool

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.fee = kwargs['fee']
        self.icon_id = kwargs['icon_id']
        self.pickup_fee = kwargs['pickup_fee']
        self.shipping_fee = kwargs['shipping_fee']
        self.total_fee = kwargs['total_fee']
        self.is_pickup = kwargs['is_pickup']
    
class ItemAttributeValue(Printable):
    id: str
    text: str

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.text = kwargs['text']

class ItemAttribute(Printable):
    id: str
    text: str
    values: list[ItemAttributeValue]
    deep_facet_filterable: bool
    show_on_ui: bool

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.text = kwargs['text']
        self.values = [ItemAttributeValue(**value) for value in kwargs['values']]
        self.deep_facet_filterable = kwargs['deep_facet_filterable']
        self.show_on_ui = kwargs['show_on_ui']

class Color(Printable):
    id: str
    name: str
    rgb: str

    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        self.name = kwargs['name']
        self.rgb = kwargs['rgb']

class ItemAuction(Printable):
    id: str
    bid_deadline: str
    total_bid: str
    highest_bid: str

    def __init__(self, *args, **kwargs):
        self.id = kwargs.get('id', "")
        self.bid_deadline = get_from_kwargs(kwargs, 'bid_deadline', 'bidDeadline', 'expected_end_time')
        self.total_bid = get_from_kwargs(kwargs, 'totalBid', 'total_bids', default="0") # yes, one has an 's'. intentional
        self.highest_bid = get_from_kwargs(kwargs, 'highest_bid', 'highestBid', default="0")

class Item(Printable):
    id: str
    productURL: str
    seller: Seller
    converted_price: ConvertedPrice
    status: str
    name: str
    price: int
    description: str
    photos: list[str]
    photo_paths: list[str]
    thumbnails: list[str]
    item_category: ItemCategory
    item_category_ntiers: ItemCategoryNTiers
    parent_categories_ntiers: list[ParentCategoryNTier]
    item_condition: ItemCondition
    colors: list[Color]
    shipping_payer: ShippingPayer
    shipping_method: ShippingMethod
    shipping_from_area: ShippingFromArea
    shipping_duration: ShippingDuration
    shipping_class: ShippingClass
    num_likes: int
    num_comments: int
    registered_prices_count: int
    comments: list[str] # this is empty in test data, find out what it is
    updated: int
    created: int
    pager_id: str
    liked: bool
    checksum: str
    is_dynamic_shipping_fee: bool
    application_attributes: dict # this is empty in test data, find out what it is
    is_shop_item: str
    hash_tags: list[str] # this is empty in test data, find out what it is
    is_anonymous_shipping: bool
    is_web_visible: bool
    is_offerable: bool
    is_organizational_user: bool
    organizational_user_status: str
    is_stock_item: bool
    is_cancelable: bool
    shipped_by_worker: bool
    additional_services: list[str] # this is empty in test data, find out what it is
    has_additional_service: bool
    has_like_list: bool
    is_offerable_v2: bool
    item_attributes: list[ItemAttribute]
    is_dismissed: bool
    photo_descriptions: list[str]
    meta_title: str
    meta_subtitle: str
    auction: ItemAuction = None # this is optional, only present if the item is an auction



    def __init__(self, *args, **kwargs):
        self.id = kwargs['id']
        # self.productURL = kwargs['productURL']
        self.seller = Seller(**kwargs['seller'])
        self.converted_price = ConvertedPrice(**kwargs['converted_price'])
        self.status = kwargs['status']
        self.name = kwargs['name']
        self.price = kwargs['price']
        self.description = kwargs['description']
        self.photos = kwargs['photos']
        self.photo_paths = kwargs['photo_paths']
        self.thumbnails = kwargs['thumbnails']
        self.item_category = ItemCategory(**kwargs['item_category'])
        self.item_category_ntiers = ItemCategoryNTiers(**kwargs['item_category_ntiers'])
        self.parent_categories_ntiers = [ParentCategoryNTier(**parent) for parent in kwargs['parent_categories_ntiers']]
        self.item_condition = ItemCondition(**kwargs['item_condition'])
        self.colors = [Color(**color) for color in kwargs['colors']] if 'colors' in kwargs else []
        self.shipping_payer = ShippingPayer(**kwargs['shipping_payer'])
        self.shipping_method = ShippingMethod(**kwargs['shipping_method'])
        self.shipping_from_area = ShippingFromArea(**kwargs['shipping_from_area'])
        self.shipping_duration = ShippingDuration(**kwargs['shipping_duration'])
        self.shipping_class = ShippingClass(**kwargs['shipping_class'])
        self.num_likes = kwargs['num_likes']
        self.num_comments = kwargs['num_comments']
        self.registered_prices_count = kwargs['registered_prices_count']
        # self.comments = kwargs['comments']
        self.comments = []
        self.updated = kwargs['updated']
        self.created = kwargs['created']
        self.pager_id = kwargs['pager_id']
        self.liked = kwargs['liked']
        self.checksum = kwargs['checksum']
        self.is_dynamic_shipping_fee = kwargs['is_dynamic_shipping_fee']
        # self.application_attributes = kwargs['application_attributes']
        self.application_attributes = {}
        self.is_shop_item = kwargs['is_shop_item']
        # self.hash_tags = kwargs['hash_tags']
        self.hash_tags = []
        self.is_anonymous_shipping = kwargs['is_anonymous_shipping']
        self.is_web_visible = kwargs['is_web_visible']
        self.is_offerable = kwargs['is_offerable']
        self.is_organizational_user = kwargs['is_organizational_user']
        self.organizational_user_status = kwargs['organizational_user_status']
        self.is_stock_item = kwargs['is_stock_item']
        self.is_cancelable = kwargs['is_cancelable']
        self.shipped_by_worker = kwargs['shipped_by_worker']
        # self.additional_services = kwargs['additional_services']
        self.additional_services = []
        self.has_additional_service = kwargs['has_additional_service']
        self.has_like_list = kwargs['has_like_list']
        self.is_offerable_v2 = kwargs['is_offerable_v2']
        self.item_attributes = [ItemAttribute(**item) for item in kwargs['item_attributes']]  if 'item_attributes' in kwargs else []
        self.is_dismissed = kwargs['is_dismissed']
        self.photo_descriptions = kwargs['photo_descriptions'] if 'photo_descriptions' in kwargs else []
        self.meta_title = kwargs['meta_title']
        self.meta_subtitle = kwargs['meta_subtitle']
        # this is optional, only present if the item is an auction
        if "auction_info" in kwargs and kwargs["auction_info"] is not None:
            self.auction = ItemAuction(**kwargs['auction_info'])
        else:
            self.auction = None



import setuptools

with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='mercari',
    version='2.2.1',
    author='marvinody',
    author_email='manny@sadpanda.moe',
    description='mercari api-like wrapper',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/marvinody/mercari/',
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.9",
    ]
)



from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, utils
from time import time
import base64
import json

def intToBytes(n):
  return n.to_bytes((n.bit_length() + 7) // 8, byteorder='big')

def intToBase64URL(n):
    return bytesToBase64URL(intToBytes(n))

def strToBase64URL(s):
    sBytes = bytes(s, 'utf-8')
    return bytesToBase64URL(sBytes)

def bytesToBase64URL(b):
    return base64.urlsafe_b64encode(b).decode('utf-8').rstrip('=')


def public_key_to_JWK(public_key):
    public_numbers = public_key.public_numbers()
    x,y = (public_numbers.x, public_numbers.y)

    return {
      "crv": "P-256", # we're hardcoding this because SECP256R1 === P-256
      "kty": "EC", # and we're always using EC, so hardcoded
      "x": intToBase64URL(x),
      "y": intToBase64URL(y),
    }
def public_key_to_Header(public_key):
    return {
      "typ": "dpop+jwt",
      "alg": "ES256", # we're hardcoding because we're not changing algo ever
      "jwk": public_key_to_JWK(public_key)
    }

# this is an util that will generate a pseudo-DPOP
# Mercari fucked up how they generate these tokens and it's not compliant with the actual spec
# But the way they screwed up makes it much easier for us.
# We can just generate random ECC keys and sign any data and as long as the keys match and signature
# is valid, then mercari doesn't care if it's a new key.
# So let's make a new key everytime so they dont blacklist any JWK public keys or something dumb
# Just make sure the method & url match the endpoint you're trying to call
# uuid SHOULD probably be a real uuid, but it doesn't seem like they're even checking this for much.
# maybe some tracking for later, so making it purely random just like the key will make it really hard
# for mercari to detect this as inorganic traffic if spread across IPs
# The biggest giveaway MAY be multiple keys/uuid being used for 1 IP.
# But then you just swap it every week or so instead and problem solved (hopefully)
def generate_DPOP(*, uuid, method, url):

    private_key = ec.generate_private_key(
        ec.SECP256R1()
    )

    # DPOP lets you specify
    payload = {
      "iat": int(time()), # 1633388418,
      "jti": uuid, # "eca01827-1724-4eea-8eb6-4532616aed9f",
      "htu": url, # "https://api.mercari.jp/search_index/search",
      "htm": method.upper(), # "GET"
    }

    # we need to give our public key to mercari, so export
    public_key = private_key.public_key()

    # create header including the JWK encoding of the key & metadata about algo we'll sign with
    header = public_key_to_Header(public_key)

    # JWT expects everything to be JSON string -> base 64 encoded
    headerString = json.dumps(header)
    payloadString = json.dumps(payload)

    # So let's encode them and add a delimiter of .
    dataToSign = f"{strToBase64URL(headerString)}.{strToBase64URL(payloadString)}"

    # Sign our resulting big string with the sk and a hardcoded algorithm
    signature = private_key.sign(
        bytes(dataToSign, 'utf-8'),
        ec.ECDSA(hashes.SHA256())
    )

    # Cryptolibrary returns an encoded structure that the webapi decryption doesn't like
    # So we need a special format. We pull out the r, s values from our DSS signature (which are ints)
    r, s = utils.decode_dss_signature(signature)
    rB, sB = intToBytes(r), intToBytes(s)

    # and then we just concat the byte representations of the r,s values & base64 the whole monster
    signatureString = bytesToBase64URL(rB + sB)

    # and concat the signature at the end
    return f"{dataToSign}.{signatureString}"

if __name__ == '__main__':
    print(generate_DPOP(uuid='DUMMY_UUID', method="GET", url="https://api.mercari.jp/search_index/search"))
